from graph_tool.all import *
import heapq
import networkx as nx
import read_graph as rg
import sys
import math
import time
import random

def sampling_graph(G,node_probs):
   #sample infected nodes then add to the infected node set
    I=0
    
    Iset=[]
    for i in range(len(node_probs)):
        w=node_probs[i]
        if (w)>random.random():
            Iset.append((i+1))    
    
    if len(Iset)==0:
        Iset.append(1)

    G.add_node(I)
    G1=G.copy()
    #printline(G)
    for node in Iset:
        for nbr in nx.all_neighbors(G, node):
            w=G[node][nbr]['weight']
            if G1.has_edge(I,nbr)==False:
                G1.add_edge(I,nbr,weight=w)
            else:
                temp=G1[I][nbr]['weight']
                G1[I][nbr]['weight']=temp+(1-temp)*w
    #remove nodes
    G=G1
    for node in Iset:
        if node in G:
            G.remove_node(node)
    G1=G
    if G1.has_edge(I,I):
        G1.remove_edge(I,I)
    return [G1,I]

def get_dominator_tree(G,I):

    domtree=nx.DiGraph()

    num_nodes=len(G)
    dicts={}
    #using graph_tool
    g=Graph()
    vlist=g.new_vertex_property("int")
    elist_weight=g.new_edge_property("double")
    for n,nbrs in G.adjacency_iter():
        for nbr, weight in nbrs.items():
            if str(n) in dicts:
                v1=int(dicts[str(n)])
            else:
                v1=g.add_vertex()
                dicts[str(n)]=str(v1)
                vlist[v1]=n

            if str(nbr) in dicts:
                v2=int(dicts[str(nbr)])
            else:
                v2=g.add_vertex()
                dicts[str(nbr)]=str(v2)
                vlist[v2]=nbr
            log_w=-math.log(weight['weight'])
            edge=g.add_edge(v1,v2)
            elist_weight[edge]=log_w
    
    #infected node
    I0=int(dicts[str(I)])

    #run shortest path to get weights for dominator tree
    dist=shortest_distance(g,I0,target=None,weights=elist_weight)
    dist_list=dist.a
    #for value in dist_list:
        #print math.exp(-value)

    
    #get dominator tree
    dom=dominator_tree(g,I0)                
    dom_list=dom.a

    for i in range(1,len(dom_list)):
        wei=0
        if g.vertex(dom_list[i])==I0:
            wei=math.exp(-dist_list[g.vertex(i)])
        else:
            wei=math.exp(dist_list[g.vertex(dom_list[i])]-dist_list[g.vertex(i)])
        domtree.add_edge(vlist[g.vertex(dom_list[i])],vlist[g.vertex(i)],weight=(wei))
        #domtree.add_edge(vlist[g.vertex(i)],vlist[g.vertex(dom_list[i])])
     
    #for n, nbrs in domtree.adjacency_iter():
     #   for nbr in nbrs:
      #      print("%d    %d" % (n,nbr))
    #printline(domtree)
    return domtree

def compute_nei_footprint(domtree,I,E_nei):
#get the footprint of I's neighbors
    #E_nei={}
     for nei in domtree[I]:
        #print nei
        pij=domtree[I][nei]['weight']
        E_nei[str(nei)]=E_nei[str(nei)]+(pij*cal_footprint(domtree,nei))
    #return E_nei
    
def cal_footprint(G,v):
    footprint=1
    if len(G[v])==0:
        footprint=1
    else:
        for nei in G[v]:
            footprint=footprint+cal_footprint(G,nei)*G[v][nei]['weight']

    return footprint

# read node probs from the graph    
def read_nodes_probs(nodes_probs_file):
    node_probs=[]
    f=open(nodes_probs_file, 'r')
    lines=f.readlines()
    for line in lines:
        [id, prob]=line.split()
        node_probs.append(eval(prob))
    f.close()
    return node_probs
        
def UVDP_sampling(filename,nodes_probs_file,K,R):
    #read graph
    print "UVDP sampling algorithm begin"
    g=rg.create_undirected_graph_IC(filename)
    #printline(g)
    
    #read nodes probability
    node_probs=read_nodes_probs(nodes_probs_file)
    
    #initialize the value
    num_nodes=len(g)
    E_nei={}
    for i in range(num_nodes):
        E_nei[str(i+1)]=0
        
    start_time=time.time()    
    
    
    for i in range(R):
        print i    
        #sampling graph G_i
        [G,I]=sampling_graph(g, node_probs)
        
        #get dominator tree
        domtree=get_dominator_tree(G,I)
        #printline(domtree)
    
        #get the benefit of I's neighbors
        compute_nei_footprint(domtree,I, E_nei)
    
    #select k nodes in dava-fast
    selected_nodes=[]
    selected_nodes=heapq.nlargest(K,E_nei,key=E_nei.get)
    
    end_time=time.time()
    cost_time=end_time-start_time
    return  [selected_nodes, cost_time] 

def printline(g):
    print "edgelist"
    for n, nbrs in g.adjacency_iter():
        for nbr,weight in nbrs.items():
            print("%d,%d,%f" %(n,nbr,weight['weight']))

def main():
    '''
    filename: edgelist
    node_probs_file: probs that nodes are intially infected
    K: the number of nodes we select
    R: how many rounds in the simulation
    '''
    filename=sys.argv[1]
    nodes_probs_file=sys.argv[2]
    K=int(sys.argv[3])
    R=int(sys.argv[4])

    #filename='edgelist'
    #nodes_probs_file='nodes_probs'
    #K=2
    #R=20

    [selected_nodes,cost_time]=UVDP_sampling(filename,nodes_probs_file,K,R)
    print cost_time
    
    save_file=filename+nodes_probs_file+'_UVDP_sample_selected_nodes'
    rg.save_selected_node(save_file,selected_nodes)
    save_file=filename+nodes_probs_file+'_UVDP_sample_cost_time'
    rg.save_cost_time(save_file,'UVDP_sample',cost_time)

if __name__=='__main__':
    main()
